#
#  DelphinOutputFile.py
#
#
#  Created by Andreas Nicolai on 10.12.09.
#  Copyright (c) 2009 IBK, TU Dresden. All rights reserved.
#

class DelphinOutputFile:
	"""Can read and interpret Delphin output files (ASCII Versions)."""
	def __init__(self):
		self.clear()

	def clear(self):
		self.headerData = dict()
		self.timePoints = []
		# holds all data as list of lists
		self.values = []
		# holds first column of data for convenient access to 2D data
		self.valueVector = []
		self.quantityDescription = ""
		# coordinate vector, holds either x, y or z coordinates of the profile cut
		self.coordinates = []
		# coordinates in x directions for all elements (order corresponds to order of elements in elements table)
		self.xCoordinates = []
		# coordinates in y directions for all elements (order corresponds to order of elements in elements table)
		self.yCoordinates = []
		# holds a string identifying the cut direction, either "X" or "Y"
		self.direction = ""
		self.filename = ""

	def readMaterialsTable(self, lines, lastLine):
		for i in range(lastLine+1, len(lines)):
			line = lines[i].strip()
			if len(line) == 0:
				continue
			if line.startswith("TABLE") or line.startswith("ELEMENTS") or line.startswith("SIDES"):
				return i
			# parse materials line... currently ignored
		raise RuntimeError("Incomplete Delphin 5 output file, missing data after MATERIALS table.")

	def readGridTable(self, lines, lastLine):
		gridX = lines[lastLine+1].split()
		gridY = lines[lastLine+2].split()
		gridZ = lines[lastLine+3].split()
		# we decide on x or y direction based on number of x values
		if len(gridZ) != 1:
			raise RuntimeError("Invalid Delphin 5 output file, must have only one z-coordinate!")

		for i in range(lastLine+4, len(lines)):
			line = lines[i].strip()
			if len(line) == 0:
				continue
			if line.startswith("TABLE") or line.startswith("ELEMENTS") or line.startswith("SIDES"):
				return i
		raise RuntimeError("Incomplete Delphin 5 output file, missing data after GRID table.")

	def readElementsTable(self, lines, lastLine):
		for i in range(lastLine+1, len(lines)):
			line = lines[i].strip()
			if len(line) == 0:
				continue
			if line.startswith("TABLE") or line.startswith("ELEMENTS") or line.startswith("SIDES"):
				return i
			# parse elements line... extract coordinates
			elementData = line.split()
			self.xCoordinates.append( float(elementData[1]) )
			self.yCoordinates.append( float(elementData[2]) )
		raise RuntimeError("Incomplete Delphin 5 output file, missing data after ELEMENTS table.")

	def readSidesTable(self, lines, lastLine):
		for i in range(lastLine+1, len(lines)):
			line = lines[i].strip()
			if len(line) == 0:
				continue
			if line.startswith("TABLE") or line.startswith("ELEMENTS") or line.startswith("SIDES"):
				return i
			# parse sides line... currently ignored
			# sidesData = line.split()
		raise RuntimeError("Incomplete Delphin 5 output file, missing data after SIDES table.")

	def read(self, fname):
		"""Reads a Delphin 5 output file (ASCII-Mode) which may have embedded geometry data.
		Returns:
		    True - on success when all data (including geometry data where needed) is present
			False - on failure, error messages are written to output stream
		"""
		self.clear()
		self.filename = fname
		# open output file and parse data
		try:
			outfile_obj = open(fname, 'r')
			lines = outfile_obj.readlines()
		except IOError:
			print "Can't read output file."
			return False
		outfile_obj.close

		if len(lines) == 0:
			print 'Empty file.'
			return False

		# check header data
		if not lines[0].find('Delphin 5 Output file'):
			print 'Not a Delphin 5 Output file.'
			return False

		# TODO : check for binary file versions

		# start parsing header until either TABLES keyword is found, or ELEMENTS/SIDES keyword is found
		for i in range(len(lines)):
			lastLine = i
			line = lines[i].strip()
			# skip empty lines
			if len(line) == 0:
				continue
			# check for end-of-header
			tokens = line.split('=')
			keyword = tokens[0].strip(" #\t")
			if len(tokens) == 1:
				if keyword.startswith("TABLE"):
					break
				# must be a comment line without keyword-value pair, just skip this
				continue
			if keyword.startswith("ELEMENTS") or keyword.startswith("SIDES"):
				break
			# parse header types
			value = tokens[1].strip()
			self.headerData[keyword] = value

		# parse embedded header, if present
		if keyword.startswith("TABLE"):
			while True:
				# parse all tables
				tableNameTokens = keyword.split()
				if len(tableNameTokens) < 2:
					raise RuntimeError('Incomplete/invalid Delphin 5 Output file, error in table definition, line:\n{}'.format(keyword))
				tableName = tableNameTokens[1].strip()
				if tableName == "MATERIALS":
					lastLine = self.readMaterialsTable(lines, lastLine)
				elif tableName == "GRID":
					lastLine = self.readGridTable(lines, lastLine)
				elif tableName == "ELEMENT_GEOMETRY":
					lastLine = self.readElementsTable(lines, lastLine)
				elif tableName == "SIDES_GEOMETRY":
					lastLine = self.readSidesTable(lines, lastLine)
				# continue reading until
				headerEnd = 0
				for i in range(lastLine, len(lines)):
					if lines[i].startswith("TABLE"):
						break
					if lines[i].startswith("ELEMENTS") or lines[i].startswith("SIDES"):
						headerEnd = i
						break
				if headerEnd == i:
					break # header end found
				# read next table
				keyword = lines[lastLine].strip()
		else:
			headerEnd = lastLine

		if headerEnd == 0:
			print 'Incomplete Delphin 5 Output file, missing "ELEMENTS" or "SIDES" definition.'
			return False

		try:

			# extract number of sides/elements
			numberString = lines[headerEnd]
			numbers = numberString.split("=")
			numbers = numbers[1].split()
			numberCount = len(numbers)

			# process selected elements and determine numbering direction and populate
			# the coordinates (cut) vector
			if numberCount > 1:
				# check if we have a FIELD file, currently sides files with multiple values are not supported
				if not self.headerData.has_key("TYPE"):
					print 'Incomplete/invalid Delphin 5 Output file header, missing TYPE keyword.'
					return False
				if self.headerData["TYPE"] != "FIELD":
					print 'Currently on Delphin 5 Output files of type FIELD are supported.'
					return False

				for i in range(numberCount):
					elementIndex = int(numbers[i])
					x = self.xCoordinates[elementIndex]
					y = self.yCoordinates[elementIndex]
					if self.direction == "X":
						self.coordinates.append(x)
					elif self.direction == "Y":
						self.coordinates.append(y)
					else:
						if i == 0:
							lastX = x
							lastY = y
						elif i == 1:
							# if we have read the second value
							deltaX = x-lastX
							if deltaX != 0:
								self.direction = "X"
								self.coordinates.append(lastX)
								self.coordinates.append(x)
							else:
								self.direction = "Y"
								self.coordinates.append(lastY)
								self.coordinates.append(y)
						else:
							assert(True)

			headerEnd = headerEnd + 1

			self.timePoints = []
			self.values = []
			lines = lines[headerEnd:]
			for line in lines:
				# parse values
				vals = line.split()
				if len(vals) == 0:
					# skip empty lines
					pass
				else:
					self.timePoints.append(float(vals[0]))
					self.valueVector.append(float(vals[1]))
					values = []
					for i in range(1, len(vals)):
						values.append(float(vals[i]))
					self.values.append(values)

			# if we have at least one output value, return true
			return (len(self.values) > 0)

		# catch raised exceptions
		except RuntimeError as err:
			print err.message
			return False

	def valuesAt(self, timeCutValue):
		# check if timeCutValue is present in list of time points
		timeCutIndex = -1
		for i in range(len(self.timePoints)):
			if self.timePoints[i] == timeCutValue:
				timeCutIndex = i
				break
		if timeCutIndex == -1:
			raise RuntimeError("Time cut value {0} does not match any output time points.".format(timeCutValue) )
		valueVector = self.values[timeCutIndex]
		if len(valueVector) != len(self.coordinates):
			raise RuntimeError("{1} values found at time cut value {0}, but only {2} coordinates/elements "
			                   "specified in elements table.".format(timeCutValue, len(valueVector), len(self.coordinates) ) )
		return valueVector
