#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Provides scripts and classes to generate and run solver simulation test cases.
"""

__author__ = "Andreas Nicolai"
__email__ = "andreas.nicolai@tu-dresden.de"
__copyright__ = "Copyright 2014, IBK, TU Dresden"
__license__ = "LGPL"
__version__ = "0.2.0"

import optparse
import os
import platform
import subprocess
import fnmatch

from simulate import *
from DelphinOutputFile import *
from Delphin6OutputFile import *
from Delphin6GeoFileManager import *


class SolverTestJob(SimJob):
	"""Simulation job specialized to execute an IBK solver job and
	show errors in the log file.
	Expects standard output paths and log file namings to work.
	"""

	def __init__(self, solverPath, projectFile):
		"""Constructs a SimJob instance and initializes the projectFile variable"""
		SimJob.__init__(self,solverPath)
		self.projectFile = projectFile

	def preproc_step(self):
		"""Called by the job manager to print the currently processed project file
		just before simulation starts.
		"""
		print self.projectFile + " ..."

	def printLastError(self):
		"""Called by the job manager just after the simulation has finished and shows,
		in case of error, the error log.
		"""
		(path,extension) = os.path.splitext(self.projectFile)
		# generate path to screenlog file
		screenlogfile = path + '/log/screenlog.txt'
		print '\033[91m'
		# open and read screenlog
		try:
			fobj = open(screenlogfile, 'r')
			for line in fobj:
				if "[Error   ]" in line and len(line) > 34:
					print "  {}".format(line[31:-1])
		except IOError:
			print "  Error reading screenlog file '{}'".format(screenlogfile)
		print '\033[0m'


class SolverTestsGenerator:
	"""Convenience function to construct simulation objects that uses an IBK solver as simulator.
	"""
	def __init__(self, solverPath, testinit):
		self.defaultCmdArgs = ["-x"]
		if testinit:
			self.defaultCmdArgs.append( "--verbosity-level=0" )
			self.defaultCmdArgs.append( "--test-init" )
		else:
			self.defaultCmdArgs.append( "--verbosity-level=0" )
		self.solverPath = solverPath

	def generateTestJob(self, project):
		job = SolverTestJob([self.solverPath] + self.defaultCmdArgs + [project], project)
		return job


def parse_default_test_args(defaultNumberOfThreads, defaultSolverPath):
	"""Parses default command line arguments for test jobs.
	Returns a tuple: (NThreads, testInitEnabled, solverPath)
	"""
	# specify standard command line arguments for test jobs
	parser = optparse.OptionParser()
	parser.add_option("", "--test-init", action="store_true", dest="testInit", default=False,
		              help="If set, the solver will be started in test-init mode.")
	parser.add_option("-n", type="int", dest="N_threads", \
		              help="Number of parallel solver tasks")

	# parse command line
	(options, args) = parser.parse_args()

	N_threads = defaultNumberOfThreads
	if options.N_threads != None:
		N_threads = int(options.N_threads)
		if N_threads <= 0:
			N_threads = 1

	# check if path to solver executable is given
	solverPath = ""
	if len(args) > 0:
		solverPath = args[0]
	else:
		solverPath = defaultSolverPath

	return (N_threads, options.testInit, solverPath)



def run_test_jobs(solverPath, solverName, projectFileExtension, testInit, numberOfThreads):
	"""This is a test generator and execution function.
	Example:
	>>> DEFAULT_SOLVER_PATH    = "../../bin/release/NandradSolver"
	>>> SOLVER_NAME            = "NandradSolver"
	>>> PROJECT_FILE_EXTENSION = "nandrad"
	>>> # number of threads to use for parallel test execution
	>>> N_THREADS              = 1
	>>> TEST_INIT              = true
	>>> # finally run tests
	>>> result = run_test_jobs(DEFAULT_SOLVER_PATH, SOLVER_NAME, PROJECT_FILE_EXTENSION, TEST_INIT, N_THREADS)
	"""

	if not os.path.isfile(solverPath):
		print "{} executable '{}' not found.".format(solverName, solverPath)
		exit(1)
	print "Using {} '{}'".format(solverName, solverPath)

	# grab all project files recursively from current directory
	matches = []
	for root, dirnames, filenames in os.walk('.'):
		for filename in fnmatch.filter(filenames, '*.{}'.format(projectFileExtension)):
			matches.append(os.path.join(root, filename))
	matches.sort()

	jobGenerator = SolverTestsGenerator(solverPath, testInit)
	jobs = []
	for fn in matches:
		jobs.append( jobGenerator.generateTestJob(fn) )

	# create the sim runner class
	simRunner = SimRunner(numberOfThreads)
	simRunner.run(jobs)

	# finally process all finished jobs and for each job with exit code != 0
	# print the fail line
	print "\n** Run Results **"

	jobCount = len(simRunner.processed_jobs)
	failCount = 0
	for job in simRunner.processed_jobs:
		if job.return_code != 0:
			print "{} : fail".format(job.cmdline[-1])
			job.printLastError()
			failCount = failCount + 1
		else:
			print "{} : pass".format(job.cmdline[-1])


	print "*** Total/Successful/Failed: {}/{}/{}".format(jobCount, jobCount-failCount, failCount)
	checkFailCount = 0
	if not testInit:
		print "\n** Check Results **"
		for fn in matches:
			# generate check script path
			(projectBasePath, ext) = os.path.splitext(fn)
			checkScript = projectBasePath + "-check.py"
			# if it does not exist, mark test as failed and go to next
			if not os.path.exists(checkScript):
				print 'Missing check script {}, test marked as failed.'.format(checkScript)
				print '{} : fail'.format(fn)
				checkFailCount = checkFailCount + 1
				continue

			# run test script
			try:
				if platform.system() == "Windows":
					cmdline = ["pythonw.exe", checkScript, fn]
					solver_process = subprocess.Popen(cmdline, creationflags=subprocess.CREATE_NEW_CONSOLE)
				else:
					cmdline = ["python", checkScript, fn]
					solver_process = subprocess.Popen(cmdline)
				# run simulation, a successful simulation should return 0
				retcode = solver_process.wait()
				if retcode != 0:
					print '{} : fail'.format(fn)
					checkFailCount = checkFailCount + 1
				else:
					print '{} : pass'.format(fn)
			except Exception as ex:
				print ex.message
				print '{} : fail'.format(fn)
				checkFailCount = checkFailCount + 1


		print "*** Total/Successful/Failed: {}/{}/{}".format(jobCount, jobCount-checkFailCount, checkFailCount)

	if failCount == 0 and checkFailCount == 0:
		return 0
	else:
		return 1

class SolverTestJob(SimJob):
	"""Simulation job specialized to execute an IBK solver job and
	show errors in the log file.
	Expects standard output paths and log file namings to work.
	"""

	def __init__(self, solverPath, projectFile):
		"""Constructs a SimJob instance and initializes the projectFile variable"""
		SimJob.__init__(self,solverPath)
		self.projectFile = projectFile

	def preproc_step(self):
		"""Called by the job manager to print the currently processed project file
		just before simulation starts.
		"""
		print self.projectFile + " ..."

	def printLastError(self):
		"""Called by the job manager just after the simulation has finished and shows,
		in case of error, the error log.
		"""
		(path,extension) = os.path.splitext(self.projectFile)
		# generate path to screenlog file
		screenlogfile = path + '/log/screenlog.txt'
		print '\033[91m'
		# open and read screenlog
		try:
			fobj = open(screenlogfile, 'r')
			for line in fobj:
				if "[Error   ]" in line and len(line) > 34:
					print "  {}".format(line[31:-1])
		except IOError:
			print "  Error reading screenlog file '{}'".format(screenlogfile)
		print '\033[0m'


class SolverTestsGenerator:
	"""Convenience function to construct simulation objects that uses an IBK solver as simulator.
	"""
	def __init__(self, solverPath, testinit):
		self.defaultCmdArgs = ["-x"]
		if testinit:
			self.defaultCmdArgs.append( "--verbosity-level=0" )
			self.defaultCmdArgs.append( "--test-init" )
		else:
			self.defaultCmdArgs.append( "--verbosity-level=0" )
		self.solverPath = solverPath

	def generateTestJob(self, project):
		job = SolverTestJob([self.solverPath] + self.defaultCmdArgs + [project], project)
		return job


def parse_default_test_args(defaultNumberOfThreads, defaultSolverPath):
	"""Parses default command line arguments for test jobs.
	Returns a tuple: (NThreads, testInitEnabled, solverPath)
	"""
	# specify standard command line arguments for test jobs
	parser = optparse.OptionParser()
	parser.add_option("", "--test-init", action="store_true", dest="testInit", default=False,
		              help="If set, the solver will be started in test-init mode.")
	parser.add_option("-n", type="int", dest="N_threads", \
		              help="Number of parallel solver tasks")

	# parse command line
	(options, args) = parser.parse_args()

	N_threads = defaultNumberOfThreads
	if options.N_threads != None:
		N_threads = int(options.N_threads)
		if N_threads <= 0:
			N_threads = 1

	# check if path to solver executable is given
	solverPath = ""
	if len(args) > 0:
		solverPath = args[0]
	else:
		solverPath = defaultSolverPath

	return (N_threads, options.testInit, solverPath)



def eval_tests(projectFileExtension):
	"""This is a test generator and execution function.
	Example:
	>>> DEFAULT_SOLVER_PATH    = "../../bin/release/NandradSolver"
	>>> SOLVER_NAME            = "NandradSolver"
	>>> PROJECT_FILE_EXTENSION = "nandrad"
	>>> # number of threads to use for parallel test execution
	>>> N_THREADS              = 1
	>>> TEST_INIT              = true
	>>> # finally run tests
	>>> result = run_test_jobs(DEFAULT_SOLVER_PATH, SOLVER_NAME, PROJECT_FILE_EXTENSION, TEST_INIT, N_THREADS)
	"""


	# grab all project files recursively from current directory
	matches = []
	for root, dirnames, filenames in os.walk('.'):
		for filename in fnmatch.filter(filenames, '*.{}'.format(projectFileExtension)):
			matches.append(os.path.join(root, filename))
	matches.sort()

	jobCount = len(matches)
	checkFailCount = 0
	print "\n** Check Results **"
	for fn in matches:
		# generate check script path
		(projectBasePath, ext) = os.path.splitext(fn)
		checkScript = projectBasePath + "-check.py"
		# if it does not exist, mark test as failed and go to next
		if not os.path.exists(checkScript):
			print 'Missing check script {}, test marked as failed.'.format(checkScript)
			print '{} : fail'.format(fn)
			checkFailCount = checkFailCount + 1
			continue

		# run test script
		try:
			if platform.system() == "Windows":
				cmdline = ["pythonw.exe", checkScript, fn]
				solver_process = subprocess.Popen(cmdline, creationflags=subprocess.CREATE_NEW_CONSOLE)
			else:
				cmdline = ["python", checkScript, fn]
				solver_process = subprocess.Popen(cmdline)
			# run simulation, a successful simulation should return 0
			retcode = solver_process.wait()
			if retcode != 0:
				print '{} : fail'.format(fn)
				checkFailCount = checkFailCount + 1
			else:
				print '{} : pass'.format(fn)
		except Exception as ex:
			print ex.message
			print '{} : fail'.format(fn)
			checkFailCount = checkFailCount + 1


	print "*** Total/Successful/Failed: {}/{}/{}".format(jobCount, jobCount-checkFailCount, checkFailCount)

	if checkFailCount == 0:
		return 0
	else:
		return 1
