#!/usr/bin/python
# -*- coding: utf-8 -*-

import shutil
import math
import os
import numpy as np
from IBK import Delphin6OutputFile
from IBK import Delphin6GeoFileManager
from IBK import simulate


### Constants and Input #########################

# Define path to DELPHIN Solver executable, preset assumes a standard Windows installation 
DELPHIN_SOLVER = r'C:\Program Files\IBK\Delphin 6.0\DelphinSolver.exe'

# Define DELPHIN project file (xxx.d6p)
DELPHIN_PROJECT_FILE = 'eisspeicher.d6p'

# Define output directory for this simulation run
DELPHIN_OUTPUT_DIR = 'eisspeicher'

# Define the output file that we want to change and use for the input of the next iteration
# Specify here the path to the DELPHIN output file without extension (so that _xxx.d6o is appended)
DELPHIN_OUTPUT_FILE = DELPHIN_OUTPUT_DIR+'/results/t_out'      

# Define if convergence check shall be used
# If CHECK_FOR_CONVERGENCE is True, we check the last value in the output file and compare it to the previous run,
# and if the difference is smaller than CONVERGENCE_THRESHOLD, the iteration is stopped.
CHECK_FOR_CONVERGENCE = True
# Define convergence threshold in DegC, to check convergence and finish iteration if under threshold
CONVERGENCE_THRESHOLD = 0.01     
# Define max number of iterations
MAX_ITERATIONS = 50     

#######################################

DELPHIN_PROJECT_NAME = DELPHIN_PROJECT_FILE[:-4]

# hack, only for testing 
if not os.path.exists(DELPHIN_SOLVER):
	DELPHIN_SOLVER = '/home/ghorwin/svn/Delphin6_trunk/bin/release/DelphinSolver' # path to solver


geoManager = Delphin6GeoFileManager()


def run_DELPHIN_model(t, Tin, iter):
	"""
	Runs the DELPHIN model using current set of input data

	Arguments:
	- Tin Vector with inlet temperatures in hourly intervals
	
	Upon success, the function returns a touple with the time-series of 
	computed outlet temperatures.
	"""
	
	# 1. Generate climate data file
	#    On first iteration Tin is empty and the original T_in.csv, copied
	#    from T_in_0.csv is used
	if len(Tin) > 0:
		fobj = open('T_in.csv', 'w')
		fobj.write('Time [d]\tTemperature [C]\n')
		for i in range(len(Tin)):
			fobj.write('{}\t{}\n'.format(t[i], Tin[i]))
		fobj.close()
		del fobj


	# 2. Start simulation
	job = simulate.SimJob([DELPHIN_SOLVER,'-x','--verbosity-level=1','-o='+DELPHIN_OUTPUT_DIR,DELPHIN_PROJECT_FILE])
	if (job.run() != 0):
		print "Error executing simulation"
		exit(1)
	del job

	# create backup copy of output temperatures for convergence analysis
	shutil.copyfile(DELPHIN_OUTPUT_FILE + '.d6o',
	                DELPHIN_OUTPUT_FILE + '_{:0>3d}.d6o'.format(iter))

	# 3. Read results
	#    Time points are expected in hours [h] !!!
	#    Values are expected in degC [C] !!!
	dout = Delphin6OutputFile()
	dout.read(DELPHIN_OUTPUT_FILE + '.d6o', geoManager)

	x = np.asarray(dout.timePoints)
	y = np.asarray(dout.valueVector)
	del dout
	
	return (x,y)
	

def run_HEAT_EXCHANGER_model(t, T):
	"""
	Simple heat pump model

	Arguments:
	
	* t - Expects argument t to be a time series in [h]
	* T - time series of temperatures in [C] for the corresponding time points in vector t
	""" 

	# defines the maximum reduction of temperature between outlet and inlet
	DELTA_T = -2     

	# process all time points in the input time series
	for i in range(len(T)):
		# Apply control algorithm:
		# - heat pump only operates between 06:00 and 18:00
		# - above 0 degree C a constant delta is applied
		#    T[i] += DELTA_T
		# - below -5 degree C heat pump is turned off
		# - linear decrease of cooling load between 0 and -5
		#    T[i] += delta(T[i])
		hour_of_day = round((t[i] * 24.0) % 24, 0)
		if 6 <= hour_of_day <= 18:
			t_in = T[i]
			if t_in < -5:
				delta = 0
			if t_in > 0:
				delta = DELTA_T
			delta = (- 0.4 * t_in + DELTA_T)
			T[i] = T[i] + delta # modify temperature for this time point
	return (t,T)
	

def gauss_seidel_iteration(t, Tin):
	# run iterations
	for iter in range(MAX_ITERATIONS):
		if iter == 1:
			convMonitor = open(DELPHIN_OUTPUT_DIR + "/log/conv_GS.csv", 'w')
		print "{: >3d}. Iteration starting".format(iter+1)
		
		# store current results for convergence check
		tLast = t
		TinLast = Tin

		# run DELPHIN simulation model
		(t1, Tout) = run_DELPHIN_model(t, Tin, iter)

		# run equipment model
		(t, Tin) = run_HEAT_EXCHANGER_model(t1, Tout)
	
		# if convergence Threshold is reached, stop iteration
		if CHECK_FOR_CONVERGENCE:
			if iter > 0:
				# simple error norm
				delta = abs(TinLast[-1] - Tin[-1])
				# WRMS norm
				weight = abs(Tin)*1e-5 + 1e-5
				normalizedT = (TinLast - Tin)/weight
				squareSum = np.sum(normalizedT**2)
				delta2 = math.sqrt(squareSum)/len(Tin)
				
				print "  ", TinLast[-1], "     - ", Tin[-1], "        = ", delta, " degC, WRMSnorm = ", delta2
				
				convMonitor.write("  {: >3d} {} {}\n".format(iter, delta, delta2))
				convMonitor.flush()
				if delta2 < 1:
					print "stop iteration"
					exit(0)
	

def gauss_jacobi_iteration(t_in, Tin, t_out, Tout):
	# run iterations
	for iter in range(MAX_ITERATIONS):
		if iter == 1:
			convMonitor = open(DELPHIN_OUTPUT_DIR + "/log/conv_GJ.csv", 'w')
		print "{: >3d}. Iteration starting".format(iter+1)
		
		# store current results for convergence check
		tLast = t_in
		TinLast = Tin

		# run DELPHIN simulation model
		(t_out2, Tout2) = run_DELPHIN_model(t_in, Tin, iter)

		# run equipment model
		(t_in2, Tin2) = run_HEAT_EXCHANGER_model(t_out, Tout)

		# update iterative quantities
		t_in = t_in2
		Tin = Tin2
		t_out = t_out2
		Tout = Tout2
	
		# if convergence Threshold is reached, stop iteration
		if CHECK_FOR_CONVERGENCE:
			if iter > 1:
				# simple error norm
				delta = abs(TinLast[-1] - Tin[-1])
				# WRMS error norm
				weight = abs(Tin)*1e-5 + 1e-5
				normalizedT = (TinLast - Tin)/weight;
				squareSum = np.sum(normalizedT**2);
				delta2 = math.sqrt(squareSum)/len(Tin)
				
				print "  ", TinLast[-1], "     - ", Tin[-1], "        = ", delta, " degC"
				
				convMonitor.write("  {: >3d} {} {}\n".format(iter, delta, delta2))
				convMonitor.flush()
				if delta2 < 1:
					print "stop iteration"
					exit(0)
					
					
# *** main program ***

# init
# - print project name to screen
print "Project:", DELPHIN_PROJECT_FILE
print '------------------------------'
# - check name convention of output file
if DELPHIN_OUTPUT_FILE[-4:] == '.d6o':
	DELPHIN_OUTPUT_FILE = DELPHIN_OUTPUT_FILE[:-4]

# start with initial T_in.ccd
shutil.copyfile('T_in_0.csv', 'T_in.csv')

# run a Gauss-Seidel-Type iteration
t_in = np.array([]) # empty initial array means "keep T_in.csv"
Tin = np.array([])

t_out = np.array([0,365*24-1])
Tout = np.array([15,15])

gauss_seidel_iteration(t_in, Tin)

exit(0)

